@@ -4,7 +4,7 @@ use strict;
 use YAML qw(LoadFile Load);
 use IO::File;
 
-our $VERSION = '0.01';
+our $VERSION = '0.03';
 
 =head1 NAME
 
@@ -264,11 +264,19 @@ sub new
         %$yaml_spec,
     };
 
+    my $config_terminator;
+    $config_terminator = $spec->{config}{config_terminator}
+        if $spec->{config} and $spec->{config}{config_terminator};
+
+    my $config_separator;
+    $config_separator = $spec->{config}{config_separator}
+        if $spec->{config} and $spec->{config}{config_separator};
+
     my $self = {
         spec       => $spec,
         debug      => $args->{debug},
-        terminator => $args->{terminator} || '~',        
-        separator  => $args->{separator}  || '*',
+        terminator => $config_terminator || $args->{terminator} || '~',
+        separator  => $config_separator  || $args->{separator}  || '*',
         error      => '',
         new_lines  => $args->{new_lines},
         truncate_null => $args->{truncate_null} || 0,
@@ -291,7 +299,8 @@ sub read_record
     my $record = { };
 
     # strip newlines if applicable
-    $string =~ s/[\r\n]//g;
+    $string =~ s/[\r\n]//g
+        unless $self->{terminator} =~ /[\r\n]/;
 
     open(my $fh, "<", \$string);
 
@@ -375,6 +384,10 @@ sub _parse_edi
     my $IN_GS = 0;
     my $IN_ST = 0;
 
+    my $IN_LOOP;
+    my $LOOP_SECTION;
+    my %LOOP_SEGMENTS;
+
     my ($current_group, $current_set, $current_record);
 
     $record->{GROUPS} = [ ]
@@ -391,12 +404,21 @@ sub _parse_edi
                 segments   => $segments,
                 type       => $type,
             });
-
             $IN_ISA = 1;
+
+            %LOOP_SEGMENTS = ();
+            $LOOP_SECTION = undef;
+            $IN_LOOP = undef;
+
         }
         elsif ($type eq 'IEA')
         {
             $IN_ISA = 0;
+
+            %LOOP_SEGMENTS = ();
+            $LOOP_SECTION = undef;
+            $IN_LOOP = undef;
+
         }
         elsif ($type eq 'GS')
         {
@@ -410,12 +432,20 @@ sub _parse_edi
 
             $IN_GS = 1;
 
+            %LOOP_SEGMENTS = ();
+            $LOOP_SECTION = undef;
+            $IN_LOOP = undef;
+
             $current_group = $new_group;
         }
         elsif ($type eq 'GE')
         {
             push @{ $record->{GROUPS} }, \%$current_group;
 
+            %LOOP_SEGMENTS = ();
+            $LOOP_SECTION = undef;
+            $IN_LOOP = undef;
+
             $IN_GS = 0;
         }
         elsif ($type eq 'ST')
@@ -428,6 +458,10 @@ sub _parse_edi
 
             $IN_ST = 1;
 
+            %LOOP_SEGMENTS = ();
+            $LOOP_SECTION = undef;
+            $IN_LOOP = undef;
+
             $current_set = $new_set;
             $current_record = $new_set;
         }
@@ -435,6 +469,10 @@ sub _parse_edi
         {
             push @{ $current_group->{SETS} }, \%$current_set;
 
+            %LOOP_SEGMENTS = ();
+            $LOOP_SECTION = undef;
+            $IN_LOOP = undef;
+
             $IN_GS = 0;
         }
         else
@@ -444,114 +482,134 @@ sub _parse_edi
 
             # parse a record
             my %segment_to_section;
+            my %loop_def;
             for my $section (keys %{ $spec->{structure} || [ ] })
             {
                 for my $segment (@{ $spec->{structure}{$section} || [ ] })
                 {
-                    $segment_to_section{$segment} = uc($section);
+                    if (ref($segment) and ref($segment) eq 'HASH')
+                    {
+                        for my $key (keys(%$segment))
+                        {
+                            $loop_def{$key} = $segment->{$key};
+                            $segment_to_section{$key} = uc($section);
+                        }
+                    }
+                    else
+                    {
+                        $segment_to_section{$segment} = uc($section);
+                    }
                 }
             }
 
             my $section  = $segment_to_section{$type};
             my $mod_record;
 
-            if (my $type_def = $spec->{segments}{uc($type)})
+            # track tree depth
+            # and dump results in loop portion?
+
             {
-                if ($section eq 'DETAIL')
-                {
-                    $current_record->{DETAIL} = [{}]
-                        unless exists $current_record->{DETAIL};
+               if (my $type_def = $spec->{segments}{uc($type)})
+               {
+                   if ($section eq 'DETAIL' or $LOOP_SECTION eq 'DETAIL')
+                   {
+                       $current_record->{DETAIL} = [{}]
+                           unless exists $current_record->{DETAIL};
 
-                    push @{ $current_record->{DETAIL} }, {}
-                        if exists($current_record->{DETAIL}->[-1]->{$type});
+                       # START THE LOOPING
+                       if ($loop_def{$type} and not $IN_LOOP)
+                       {
+                           $IN_LOOP = $type;
+                           $LOOP_SECTION = $section;
+                           %LOOP_SEGMENTS = map { $_ => 1 } @{ $loop_def{$type} || [] };
+                       }
 
-                    $current_record->{DETAIL}->[-1]->{$type} = $self->_parse_definition({
-                        definition => $type_def->{definition},
-                        segments   => $segments,
-                        type       => $type,
-                    });
-                }
-                else
-                {
-                    $current_record->{$section} = {}
-                        unless exists $current_record->{$section};
+                       # END THE LOOPING
+                       if ($IN_LOOP and not $LOOP_SEGMENTS{$type})
+                       {
+                           %LOOP_SEGMENTS = ();
+                           $LOOP_SECTION = undef;
+                           $IN_LOOP = undef;
+                       }
 
-                    $current_record->{$section}->{$type} = $self->_parse_definition({
-                        definition => $type_def->{definition},
-                        segments   => $segments,
-                        type       => $type,
-                    });
-                }
-            }
+                       if ($IN_LOOP)
+                       {
+                           $current_record->{DETAIL}->[-1]->{$IN_LOOP} = [{}]
+                               unless exists $current_record->{DETAIL}->[-1]->{$IN_LOOP};
+
+                           push @{ $current_record->{DETAIL}->[-1]->{$IN_LOOP} }, {}
+                               if exists($current_record->{DETAIL}->[-1]->{$IN_LOOP}->[-1]->{$type});
+
+                           $current_record->{DETAIL}->[-1]->{$IN_LOOP}->[-1]->{$type} = $self->_parse_definition({
+                               definition => $type_def->{definition},
+                               segments   => $segments,
+                               type       => $type,
+                           });
+                       }
+                       else
+                       {
+                           push @{ $current_record->{DETAIL} }, {}
+                               if exists($current_record->{DETAIL}->[-1]->{$type});
+
+                           $current_record->{DETAIL}->[-1]->{$type} = $self->_parse_definition({
+                               definition => $type_def->{definition},
+                               segments   => $segments,
+                               type       => $type,
+                           });
+                       }
+                   }
+                   else
+                   {
+                       $current_record->{$section} = {}
+                           unless exists $current_record->{$section};
+
+                       my $structure = $spec->{structure}{lc($section)};
+
+                       # START THE LOOPING
+                       if ($loop_def{$type} and not $IN_LOOP)
+                       {
+                           $IN_LOOP = $type;
+                           $LOOP_SECTION = $section;
+                           %LOOP_SEGMENTS = map { $_ => 1 } @{ $loop_def{$type} || [] };
+                       }
+
+                       # END THE LOOPING
+                       if ($IN_LOOP and not $LOOP_SEGMENTS{$type})
+                       {
+                           %LOOP_SEGMENTS = ();
+                           $LOOP_SECTION = undef;
+                           $IN_LOOP = undef;
+                       }
+
+                       if ($IN_LOOP)
+                       {
+                           $current_record->{$LOOP_SECTION}->{$IN_LOOP} = [{}]
+                               unless exists $current_record->{$LOOP_SECTION}->{$IN_LOOP};
+
+                           push @{ $current_record->{$LOOP_SECTION}{$IN_LOOP} }, {}
+                               if exists($current_record->{$LOOP_SECTION}->{$IN_LOOP}->[-1]->{$type});
+
+                           $current_record->{$LOOP_SECTION}->{$IN_LOOP}->[-1]->{$type} = $self->_parse_definition({
+                               definition => $type_def->{definition},
+                               segments   => $segments,
+                               type       => $type,
+                           });
+                       }
+                       else
+                       {
+                           $current_record->{$section}->{$type} = $self->_parse_definition({
+                               definition => $type_def->{definition},
+                               segments   => $segments,
+                               type       => $type,
+                           });
+                       }
+                   }
+               }
+           }
         }
     }
 }
 
-sub _parse_transaction_set
-{
-    my ($self, $params) = @_;
-
-    my $fh             = $params->{fh};
-    my $record         = $params->{record};
-    my $definition     = $params->{definition};
-    my $string         = $params->{string};
-    my $rows           = $params->{rows};
-
-    my $buffer = '';
-
-    my %segment_to_section;
-    for my $section (keys %{ $self->{spec}{structure} || [ ] })
-    {
-        for my $segment (@{ $self->{spec}{structure}{$section} || [ ] })
-        {
-            $segment_to_section{$segment} = $section;
-        }
-    }
-
-    for my $segments (@{ $rows })
-    {
-        my $type     = shift(@$segments);
-        my $section  = $segment_to_section{$type};
-        my $mod_record;
-
-        if (my $type_def = $self->{spec}->{segments}{uc($type)})
-        {
-            if ($section eq 'detail')
-            {
-                $record->{detail} = [{}]
-                    unless exists $record->{detail};
-
-                $mod_record = $record->{detail}->[-1];
-
-                if (exists($mod_record->{$type}))
-                {
-                    push @{ $record->{detail} }, {};
-                    $mod_record = $record->{detail}->[-1];
-                }
-            }
-            else
-            {
-                $record->{$section} = {}
-                    unless exists $record->{$section};
-
-                $mod_record = $record->{$section};
-            }
-
-            $self->_parse_definition({
-                record     => $mod_record,
-                definition => $type_def->{definition},
-                segments   => $segments,
-                type       => $type,
-            });
-        }
-        else
-        {
-            die __PACKAGE__ . ": `$type` not supported";
-        }
-
-    }
-}
-
 sub _write_spec
 {
     my ($self, %params) = @_;
@@ -647,7 +705,13 @@ sub _write_edi
 
         for my $set (@{ $group->{SETS} || [ ] })
         {
-            my $record_count = 2;
+            my $record_count = 1;
+
+            # you don't want to know why this exists...
+            if ($self->{spec}->{RECORD_OFFSET_COUNT})
+            {
+                $record_count = $record_count + $self->{spec}->{RECORD_OFFSET_COUNT};
+            }
 
             $set_count++;
 
@@ -672,12 +736,33 @@ sub _write_edi
             # process set header
             for my $section (@{ $spec->{structure}{header} || [ ] })
             {
-                $record_count++;
-                print $fh $self->_write_spec(
-                    type     => $section,
-                    type_def => $spec->{segments}{$section},
-                    record   => $set->{HEADER}{$section},
-                );
+                if (ref($section) and ref($section) eq 'HASH')
+                {
+                    my ($loop_name) = keys(%$section);
+                    my $loop_structure = $section->{$loop_name};
+
+                    for my $record (@{ $set->{HEADER}{$loop_name} || [] })
+                    {
+                        for my $structure (@{ $loop_structure || []})
+                        {
+                            $record_count++;
+                            print $fh $self->_write_spec(
+                                type     => $structure,
+                                type_def => $spec->{segments}{$structure},
+                                record   => $record->{$structure},
+                            );
+                        }
+                    }
+                }
+                else
+                {
+                    $record_count++;
+                    print $fh $self->_write_spec(
+                        type     => $section,
+                        type_def => $spec->{segments}{$section},
+                        record   => $set->{HEADER}{$section},
+                    );
+                }
             }
 
             # process set details
@@ -685,12 +770,33 @@ sub _write_edi
             {
                 for my $section (@{ $spec->{structure}{detail} || [ ] })
                 {
-                    $record_count++;
-                    print $fh $self->_write_spec(
-                        type     => $section,
-                        type_def => $spec->{segments}{$section},
-                        record   => $detail->{$section},
-                    );
+                    if (ref($section) and ref($section) eq 'HASH')
+                    {
+                        my ($loop_name) = keys(%$section);
+                        my $loop_structure = $section->{$loop_name};
+
+                        for my $sub_record (@{ $detail->{$loop_name} || [] })
+                        {
+                            for my $structure (@{ $loop_structure || []})
+                            {
+                                $record_count++;
+                                print $fh $self->_write_spec(
+                                    type     => $structure,
+                                    type_def => $spec->{segments}{$structure},
+                                    record   => $sub_record->{$structure},
+                                );
+                            }
+                        }
+                    }
+                    else
+                    {
+                        $record_count++;
+                        print $fh $self->_write_spec(
+                            type     => $section,
+                            type_def => $spec->{segments}{$section},
+                            record   => $detail->{$section},
+                        );
+                    }
                 }
             }
             
@@ -698,12 +804,33 @@ sub _write_edi
             # process set footer
             for my $section (@{ $spec->{structure}{footer} || [ ] })
             {
-                $record_count++;
-                print $fh $self->_write_spec(
-                    type     => $section,
-                    type_def => $spec->{segments}{$section},
-                    record   => $set->{FOOTER}{$section},
-                );
+                if (ref($section) and ref($section) eq 'HASH')
+                {
+                    my ($loop_name) = keys(%$section);
+                    my $loop_structure = $section->{$loop_name};
+
+                    for my $record (@{ $set->{FOOTER}{$loop_name} || [] })
+                    {
+                        for my $structure (@{ $loop_structure || []})
+                        {
+                            $record_count++;
+                            print $fh $self->_write_spec(
+                                type     => $structure,
+                                type_def => $spec->{segments}{$structure},
+                                record   => $record->{$structure},
+                            );
+                        }
+                    }
+                }
+                else
+                {
+                    $record_count++;
+                    print $fh $self->_write_spec(
+                        type     => $section,
+                        type_def => $spec->{segments}{$section},
+                        record   => $set->{FOOTER}{$section},
+                    );
+                }
             }
 
             ######
@@ -718,8 +845,6 @@ sub _write_edi
                     control_number => $set->{control_number},
                 },
             );
-
-            
         }
 
         # process GE line
@@ -741,84 +866,7 @@ sub _write_edi
             control_number => $record->{ISA}{control_number},
             total          => $group_count,
         },
-    );
-    
-    # write details
-    for my $detail ( @{ $record->{detail} || [ ] } )
-    {
-        for my $section (@{ $self->{spec}{structure}{detail} || [ ] })
-        {
-            if (my $type_def = $self->{spec}->{segments}{uc($section)})
-            {
-                my @line = ($section);
-
-                for my $def (@{ $type_def->{definition} || [ ] })
-                {
-                    my $value = exists $detail->{$section}{$def->{name}} ?
-                        $detail->{$section}{$def->{name}} :
-                        $def->{value};
-
-                    $def->{bytes} ||= '';
-
-                    # deal with minimum
-                    $def->{bytes} = $def->{min}
-                        if $value ne '' and not($def->{bytes}) and $def->{min} and length($value) < $def->{min};
-
-                    $def->{bytes} = '-' . $def->{bytes}
-                        if $def->{bytes};
-
-                    my $format = $def->{format} || "\%$def->{bytes}s";
-
-                    # deal with maximum limits
-                    $value = substr($value, 0, $def->{max})
-                        if $def->{max};
-
-                    push @line, sprintf($format, $value);
-                }
-
-                print $fh join($sep_val, @line);
-                print $fh $term_val;
-                print $fh "\n" if $self->{new_lines};
-            }
-        }
-    }
-
-    # write footer
-    for my $section (@{ $self->{spec}{structure}{footer} || [ ] })
-    {
-        if (my $type_def = $self->{spec}->{segments}{uc($section)})
-        {
-            my @line = ($section);
-
-            for my $def (@{ $type_def->{definition} || [ ] })
-            {
-                my $value = exists $record->{footer}{$section}{$def->{name}} ?
-                    $record->{footer}{$section}{$def->{name}} :
-                    $def->{value};
-
-                $def->{bytes} ||= '';
-
-                # deal with minimum
-                $def->{bytes} = $def->{min}
-                    if $value ne '' and not($def->{bytes}) and $def->{min} and length($value) < $def->{min};
-
-                $def->{bytes} = '-' . $def->{bytes}
-                    if $def->{bytes};
-
-                my $format = $def->{format} || "\%$def->{bytes}s";
-
-                # deal with maximum limits
-                $value = substr($value, 0, $def->{max})
-                    if $def->{max};
-
-                push @line, sprintf($format, $value);
-            }
-
-            print $fh join($sep_val, @line);
-            print $fh $term_val;
-            print $fh "\n" if $self->{new_lines};
-        }
-    }
+    );    
 }
 
 =head1 EXAMPLES
@@ -958,9 +1006,13 @@ sub _write_edi
  my $record = $x12->read_record($string);
  my $out = $x12->write_record($record);
 
+=head2 LOOPS
+
+ Both implicit and explicit loop segments are also supported by this module.  Please review the loops test for an example.
+
 =head1 HISTORY
 
-This module was authored for Bizowie's L<erp software|http://bizowie.com/>.
+This module was authored for L<Bizowie|http://bizowie.com/>.
 
 =head1 AUTHOR
 
@@ -968,7 +1020,7 @@ Bizowie
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2014 Bizowie
+Copyright (C) 2014, 2015, 2016 Bizowie
 
 This library is free software; you can redistribute it and/or modify it under the same terms as Perl itself, either Perl version 5.14.2 or, at your option, any later version of Perl 5 you may have available.
 
